// copyright (c) 2000 Fredrik Sjoholm <fredrik@sjoholm.com>

// rewrite
// Usage: rewrite [-v] file [file2] "->" "cat|cat|cat" "->" [rename_file_to [rename_file2_to]]

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>
#include <assert.h>
#include "vbuf.h"


#define MMAPCHUNKSIZE	(4*1024*1024)
#define BUFSIZE		((3*MMAPCHUNKSIZE)/2)

#define BSEGSIZE	32768
#define BMINSEG		64

typedef struct {
  int	wpos;	// filling the buffer at this index
  int	rpos;	// draiing the buffer at this index
  char*	data;
  int	size;	// the buffer size data[size]
  int	fpos;	// buffer starts this number of bytes into file (if buffer is mmaped)
  int	fsize;	// file size (if buffer is mmaped)

} Buffer;

// Buffer macros
#define Bempty(b) ((b).wpos == (b).rpos)		// is buffer empty? (if we got 1 seg, wpos and rpos are still valid)
#define Bfull(b) ((b).wpos == (b).size)			// is buffer full? (and no more seg headers, if using segs)
#define Bdone(b) ((b).fpos == (b).fsize)		// is an mmaped buffer done?
#define Bwritedisk(rb,wbv) ( \
	( ((wbv).drain.tot < (rb).fpos) && (!VBempty(wbv)) ) || \
	( Bdone(rb) && (!VBempty(wbv)) )  )		// are we ready to write to disk?

#define Breadpipe(rb,wbv) (!Bwritedisk(rb,wbv))		// should we read from pipe?
#define Bclear(b) (b).wpos = (b).rpos = (b).fpos = (b).fsize = 0

#define Biggest(a,b) ((a) > (b) ? (a) : (b))

void show_usage_and_exit (char *hint);
char** duprange (char** args, int n);
int bipopen (const char* command, int suck[2], int blow[2]);
int balloc (Buffer* b, int new_size);
int is_pipeline_fucked (int waitforit);
int setproctitle (const char* fmt, ...);
void do_one_file (int fn, int islast);				// process one file thru open pipe
void bail_out(int fn);
char** run_sed (char* exp, char** input, int args);




char** argvp;			// copy of argv
char* argv0;
char *pipecmd;
char **filenames, **newnames;
int pid;			// pid of child pipeline
int verbose;

Vbuf wbv;			// socket-to-file buffer
Buffer rb;			// file-to-socket buffer (mmaped)
int wp, rp;			// write-pipe, read-pipe


int main(int argc, char **argv)
{
  int pipe0[2];	// [0]: read, [1]: write
  int pipe1[2];
  int ac;
  int start;
  int nfiles = 0;
  char *sedexp = 0;
  int goodargs = 0;
  int i;

  argvp = argv;
  argv0 = strrchr(argv[0],'/');
  if (argv0) ++argv0; else argv0 = argv[0];
  argv0 = strdup(argv0);

  for (ac = 1, start = 0; ac < argc; ++ac) {

    int islast = ac == argc-1;

    if (islast) {
      if (!nfiles) show_usage_and_exit("No input files");
      filenames = duprange (argv+start, nfiles);
      if (sedexp) newnames = run_sed (sedexp, filenames, nfiles);
      pipecmd = strdup (argv[ac]);
      goodargs = 1;
    }
    else if (!strcmp("-v",argv[ac])) {
      verbose = 1;
    }
    else if (!strcmp("-r",argv[ac])) {
      sedexp = argv[++ac];
    }
    else {
      if (!start) start = ac;
      nfiles++;
    }

  }


  if (!goodargs) show_usage_and_exit(0);


  
  if (pipe(pipe0)<0 || pipe(pipe1)<0) {
    fprintf(stderr, "%s: pipe failed: %s\n", argv0, strerror(errno));
    exit(1);
  }

  pid = bipopen (pipecmd, pipe1, pipe0);
  rp = pipe0[0];
  wp = pipe1[1];



  for (i=1; i<=nfiles; ++i) {
    do_one_file (i-1, i==nfiles);
  }

  exit(0);
}



// process one file

void do_one_file (int fn, int islast)
{
  int rd, wr;
  struct stat statb;
  int cnt;
  int reading_pipe = 1;		// are we still reading (this file) from pipe?
  int writing_pipe = 1;		// are we still writing (this file) to pipe
  struct timeval timeout;
  int n;

  // open file for reading and writing
  rd = open (filenames[fn], O_RDONLY);
  if (rd < 0) {
    fprintf(stderr, "%s: failed to open \"%s\" for reading: %s\n", argv0, filenames[fn], strerror(errno));
    exit(1);
  }

  wr = open (filenames[fn], O_WRONLY);
  if (wr < 0) {
    fprintf(stderr, "%s: failed to open \"%s\" for writing: %s\n", argv0, filenames[fn], strerror(errno));
    exit(1);
  }


  bzero (&rb, sizeof(rb));
  vb_clear (&wbv);

  // now we're ready.
  // rd: read from file
  // wp: write to pipe
  // rp: read from pipe
  // wr: write to file


  if (fstat(rd,&statb) < 0) {
    fprintf (stderr, "%s: failed to stat \"%s\": %s\n", argv0, filenames[fn], strerror(errno));
    exit(1);
  }
  rb.fsize = statb.st_size;

  if (fcntl(wp,F_SETFL,O_NONBLOCK) < 0) {	// make sure we never block when writing the pipe
    fprintf (stderr, "%s: failed to make pipe nonblocking: %s\n", argv0, strerror(errno));
    exit(1);
  }


  signal (SIGPIPE, SIG_IGN);	// dont die if the pipe breaks

  // keep going while:
  //	we haven't read the entire file, or
  //	the pipe is still open for reading, or
  //	the file-write buffer is not empty

  for (cnt=0; rb.fsize && (!Bdone(rb) || reading_pipe || !VBempty(wbv)); ++cnt) {

    fd_set rdset, wrset;
    int hifd;

//    if (rb.fsize && cnt%100 == 0) {
//      setproctitle ("%s: %.0f%% %dk -> %dk", argv0,
//		(100.*(rb.fpos+rb.rpos))/rb.fsize, rb.fsize/1024, wb.fsize/1024);
//    }

    // mmap a new region of input file

    if (Bempty(rb) && rb.fpos < rb.fsize) {

      if (rb.data) munmap (rb.data, rb.size);

      rb.fpos = rb.fpos + rb.size;
      rb.size = MMAPCHUNKSIZE;
      if (rb.fpos + rb.size > rb.fsize)
	rb.size = rb.fsize - rb.fpos;

      if (rb.size) {
//printf("mmap(0,%d,PROT_READ,MAP_SHARED,%d,%d);\n", rb.size, rd, rb.fpos);
	rb.data = mmap(0, rb.size, PROT_READ, MAP_SHARED, rd, rb.fpos);
	if (rb.data == MAP_FAILED) {
	  fprintf(stderr,"%s: mmap'ing file \"%s\" failed: %s\n", argv0, filenames[fn], strerror(errno));
	  bail_out(fn);
	}
      } else {
//printf("nothing left to mmap. readbuffer done.\n");
      }
      rb.wpos = rb.size;
      rb.rpos = 0;
    }


    // see if there's any data ready to read from pipe

    hifd = -1;
    FD_ZERO(&rdset); FD_ZERO(&wrset);
    if (reading_pipe && Breadpipe(rb,wbv)) {	// reading from pipe?
      FD_SET(rp,&rdset);
      hifd = Biggest(hifd,rp);
    }
    if (writing_pipe && rb.rpos < rb.wpos) {	// writing to pipe?
      FD_SET(wp,&wrset);
      hifd = Biggest(hifd,wp);
    }
    if (Bwritedisk(rb,wbv)) {			// can we write to file? (always do if possible)
      FD_SET(wr,&wrset);
      hifd = Biggest(hifd,wr);
    }
    if (hifd < 0) {
//printf ("nothing left to do!\n");
      // nothing left to do! we can't read or write pipe, and we can't write to file
      exit(0);
    }

    timeout.tv_sec = 0;
    timeout.tv_usec = 250000;	// sleep for 250ms at the most

    n = select (hifd+1, &rdset, &wrset, 0, (islast ? 0 : &timeout));

    if (n == 0 && !writing_pipe && !Bwritedisk(rb,wbv)) {
      // we're stuck, we have to switch files (which mean some of this file's
      // date may go in the next output file. no way to tell... ;)
//printf ("Continuing with next file..\n");
      break;
    }
    else if (n < 0) {
      fprintf(stderr, "%s: select failed: %s\n", argv0, strerror(errno));
      exit(1);
    }

    // feed the pipe
    if (writing_pipe && FD_ISSET(wp,&wrset)) {
      int n = rb.wpos-rb.rpos;
      n = write (wp, rb.data+rb.rpos, n);
//printf("w"); fflush(stdout);
      if (n <= 0) {
	fprintf(stderr,"%s: failed writing to pipeline \"%s\": %s\n", argv0, pipecmd, strerror(errno));
	if (wbv.drain.tot > 0) {
	  // shit, pipe broken and we already started trashing file
	  is_pipeline_fucked(1);
	}
	bail_out(fn);

      } else {
	rb.rpos += n;
      }
      if (rb.fpos + rb.rpos == rb.fsize) {
	// we've written entire file to pipe
	writing_pipe = 0;
	if (islast) { close(wp); wp = -1; }
//printf ("\nall input written to pipe.\n");
      }
    }


    // write to file
    if (wr>=0 && FD_ISSET(wr,&wrset)) {
      int n = VBcontents(wbv);
      if (!Bdone(rb) && (wbv.fill.tot > rb.fpos))	// make sure we never write beyond our read position
	n = rb.fpos - wbv.drain.tot;

      if (wbv.drain.tot == 0) {
	// first time write, make sure pipeline didn't blow up before we trash the file
	if (is_pipeline_fucked(0)) {
	  bail_out(fn);
	}
	// protect ourselves from dangerour signals now that the file is in limbo
	signal (SIGINT, SIG_IGN);
	signal (SIGTERM, SIG_IGN);
	signal (SIGQUIT, SIG_IGN);
//printf ("protected!\n");
      }

//      n = write (wr, wb.data+wb.rpos, n);
//printf("trying to write %d\n", n);
      n = vb_write (&wbv, wr, n);
//printf("wrote %d bytes to file\n", n);
      if (n <= 0) {
	fprintf(stderr,"%s: failed writing to file \"%s\": %s\n", argv0, filenames[fn], strerror(errno));
	bail_out(fn);
      }
    }


    // read the pipe
    if (reading_pipe && FD_ISSET(rp,&rdset)) {
      int n;
      n = vb_read (&wbv, rp, BSEGSIZE);
//printf("READ %d bytes\n", n);
//printf("r"); fflush(stdout);
      if (n == 0) {
	reading_pipe = 0;
	if (islast) { close(rp); rp = -1; }		// pipe closed
      }
      else if (n < 0) {
	fprintf(stderr,"%s: failed reading from pipeline \"%s\": %s\n", argv0, pipecmd, strerror(errno));
	if (wbv.drain.tot > 0) {
	  // shit, pipe broken and we already started trashing file
	  is_pipeline_fucked(1);
	}
	bail_out(fn);
      }
    }


  }

  // did the file get smaller?
  if (wbv.drain.tot < rb.fpos) {
    if (ftruncate(wr, wbv.drain.tot) < 0) {
      fprintf (stderr, "%s: failed to truncate file \"%s\" to %ld bytes: %s\n",
		argv0, filenames[fn], wbv.drain.tot, strerror(errno));
      exit(2);
    }
  }

  if (newnames) {
    if (rename(filenames[fn],newnames[fn]) < 0) {
      fprintf (stderr, "%s: failed to rename \"%s\" to \"%s\": %s\n", argv0, filenames[fn], newnames[fn], strerror(errno));
      exit(2);
    }
  }

  if (verbose) {
    double ratio = rb.fpos ? 100.*(rb.fpos-wbv.drain.tot)/rb.fpos : 100.;
    if (newnames) {
      printf ("%s: %10d -> %10ld bytes %6.1f%% for \"%s\" -> \"%s\"\n",
		argv0, rb.fpos, wbv.drain.tot, ratio, filenames[fn], newnames[fn]);
    } else {
      printf ("%s: %10d -> %10ld bytes %6.1f%% for \"%s\"\n",
		argv0, rb.fpos, wbv.drain.tot, ratio, filenames[fn]);
    }
  }

  close(rd);
  close(wr);

}



// handle signals, while safe to do so.
// later (after we start writing to the file, most dangerous signals will be ignored.)
// since our pipeline children have their own process group,
// we need to kill them ourselves if we get SIGINT..

void sig_handler(int sig)
{
  kill (-pid, sig);
  exit(1);
}



// b		buffer structure to allocate
// new_size	grow/allocate this much for buffer
// danger	0: we're safe, 1: pray we're not gonna fail
// return	0 for failure

int balloc (Buffer* b, int new_size)
{
  if (new_size > b->size) {

    b->data = realloc (b->data, new_size);

    if (!b->data) {
      fprintf(stderr, "%s: failed to allocate %d bytes for buffer: %s\n", argv0, new_size, strerror(errno));
      bail_out(-1);
    }

    b->size = new_size;
  }
  return 1;
}



// open a pipe with descriptors on both ends.
// pass it 2 new pipes. all unnecessary descriptors will be closed
// cnild will suck from suck[0], and blow into blow[1]
// returns pid

int bipopen (const char* command, int suck[2], int blow[2])
{
  int pid;

  // it's safe to install handlers before forking, since handlers
  // will be set back to default by execvp()
  // there is no race condition! if sig_handler() is called for child,
  // it will safely kill itself (since pid is 0 in child process)
  signal (SIGINT, sig_handler);
  signal (SIGTERM, sig_handler);
  signal (SIGQUIT, sig_handler);
  // note: the parent process will later ignore these signals

  pid = fork();

  if (pid == -1) {
    fprintf(stderr, "%s: fork failed: %s\n", argv0, strerror(errno));
    exit(1);
  }

  if (pid == 0) {
    char *argv[4];
    argv[0] = "pipeline";
    argv[1] = "-c";
    argv[2] = (char*)command;
    argv[3] = 0;
    close(suck[1]);
    if (dup2(suck[0],0)<0 || close(suck[0])<0) {
      fprintf(stderr, "%s: input pipe failed: %s\n", argv0, strerror(errno));
      exit(1);
    }
    close(blow[0]);
    if (dup2(blow[1],1)<0 || close(blow[1])<0) {
      fprintf(stderr, "%s: output pipe failed: %s\n", argv0, strerror(errno));
      exit(1);
    }
    setpgid (0,0);
    execvp("/bin/sh", argv);
    fprintf(stderr, "%s: exec /bin/sh failed: %s\n", argv0, strerror(errno));
    exit(1);
  }
  else {
    // close parent left-over descriptors
    close(suck[0]);
    close(blow[1]);
  }

  return pid;
}



// waitforit	blocking wait until it dies
// return:	0 if its still alive, or it exited normally, 1 if it aborted

int is_pipeline_fucked (int waitforit)
{
  int status;
  int wpid;

  if (pid == 0) return 0;

  do {
    wpid = waitpid (pid, &status, (waitforit ? 0 : WNOHANG) );
  } while (wpid < 0);	// keep doing if we're getting signals


  // show messages if:
  //	the child is dead (we got wpid), and
  //	it did not exit normally; or it exited with a non-zero status

  if (wpid && (!WIFEXITED(status) || WEXITSTATUS(status))) {
    if (WIFSIGNALED(status)) {
      fprintf(stderr,"%s: pipeline \"%s\" killed by signal %d\n", argv0, pipecmd, WTERMSIG(status));
    } else {
      fprintf(stderr,"%s: pipeline \"%s\" exited with status %d\n", argv0, pipecmd, WEXITSTATUS(status));
    }
    pid = 0;	// we dont have any kids anymore
    return 1;
  }
  else if (wpid) {
    pid = 0;	// we dont have any kids anymore
  }

  return 0;
}


// exit, and explain what happened
void bail_out(int fn)
{
  if (fn < 0) {
    fprintf(stderr,"%s: no changes made to any files yet: Exiting safely\n", argv0);
    exit(1);
  }
  if (wbv.drain.tot > 0) {
    fprintf(stderr,"%s: %ld bytes written to file \"%s\": It's fucked, sorry\n", argv0, wbv.drain.tot, filenames[fn]);
    exit(2);
  }
  else {
    fprintf(stderr,"%s: no changes made to file \"%s\" yet: Exiting safely\n", argv0, filenames[fn]);
    exit(1);
  }
}



// hack, if sendmail can do it, we can too

int setproctitle (const char* fmt, ...)
{
  int n;
  va_list ap;
  va_start(ap, fmt);
  n = snprintf (argvp[0], 80, fmt, ap);
  va_end(ap);
  return n;
}

// strdup a section of an argument list

char** duprange (char** args, int n)
{
  char** out = calloc (n+1, sizeof(char*));
  while (n--)
    out[n] = strdup (args[n]);
  return out;
}








// parse input strings thru sed, one per line

char** run_sed (char* exp, char** input, int args)
{
  FILE *input_file;
  FILE *script_file;
  FILE *sed;
  char input_file_name[L_tmpnam];
  char script_file_name[L_tmpnam];
  char tmp[1024];
  char **output;
  int i;

  tmpnam ((char*)&input_file_name);
  tmpnam ((char*)&script_file_name);

  // write user input to file
  ///////////////////////////
  input_file = fopen (input_file_name, "w");
  if (!input_file) {
    fprintf (stderr, "%s: failed to create temp file %s: %s\n", argv0, input_file_name, strerror(errno));
    exit(1);
  }
  for (i=0; i<args; ++i) { 
    fprintf (input_file, "%s\n", input[i]);
  }
  fclose (input_file);

  // write user script to file
  ////////////////////////////
  script_file = fopen (script_file_name, "w");
  if (!script_file) {
    fprintf (stderr, "%s: failed to create temp file %s: %s\n", argv0, script_file_name, strerror(errno));
    exit(1);
  }
  fputs (exp, script_file);
  fclose (script_file);

  sprintf (tmp, "/bin/sed -f %s %s", script_file_name, input_file_name);

  // run input through /bin/sed
  /////////////////////////////
  sed = popen (tmp, "r");
  if (!sed) {
    fprintf (stderr, "%s: failed open pipe to /bin/sed: %s\n", argv0, strerror(errno));
    exit(1);
  }
  output = (char**)calloc (args+1, sizeof(char*));

  for (i=0; i<args && !feof(sed); ++i) {
    char *nl;
    if (fgets((char*)&tmp,sizeof(tmp),sed) == 0) {
      fprintf (stderr, "%s: failed to read line from sed: %s\n", argv0, strerror(errno));
      exit(1);
    }
    if (strlen(tmp) == sizeof(tmp) - 1) {
      fprintf (stderr, "%s: output filename too long: %s...\n", argv0, tmp);
      exit(1);
    }
    nl = strchr (tmp,'\n'); if (nl) *nl = 0;	// trim newlines
    output[i] = strdup(tmp);
  }

  // clean up
  ///////////
  remove (script_file_name);
  remove (input_file_name);

  i = pclose (sed);
  if (i != 0) {
    fprintf (stderr, "%s: pipe to /bin/sed failed with exit status %d\n", argv0, i);
    exit(1);
  }

  return output;
}







void show_usage_and_exit(char *hint)
{
  if (hint)
    fprintf(stderr, "%s: %s\n", argv0, hint);

  fprintf(stderr,
	"Usage: %s [-v] [-r \"s/old/new/\"] files.. \"cat|cat\"\n"
	"\n"
	"	-v		verbose output (show compression ratios)\n"
	"	s/old/new/	sed expression for renaming files\n"
	"\n"
	"	files..		input file names to rewrite through pipeline\n"
	"	\"cat|cat\"	pipeline expression to rewrite file through\n"
	"\n", argv0);
  exit(1);
}





